﻿/*
    Egaroucid Project

    @date 2021-2022
    @author Takuto Yamana (a.k.a Nyanyan)
    @license GPL-3.0 license
*/

#pragma once
#include <iostream>
#include <future>
#include "./../ai.hpp"
#include "./../umigame.hpp"
#include "function/language.hpp"
#include "function/menu.hpp"
#include "function/graph.hpp"
#include "function/opening.hpp"
#include "function/button.hpp"
#include "function/radio_button.hpp"
#include "gui_common.hpp"

#define HINT_SINGLE_TASK_N_THREAD 4

bool compare_value_cell(pair<int, int>& a, pair<int, int>& b) {
	return a.first > b.first;
}

bool compare_hint_info(Hint_info& a, Hint_info& b) {
	return a.value > b.value;
}

void draw_board(Fonts fonts, Colors colors, History_elem history_elem) {
	String coord_x = U"abcdefgh";
	for (int i = 0; i < HW; ++i) {
		fonts.font_bold(i + 1).draw(15, Arg::center(BOARD_SX - BOARD_COORD_SIZE, BOARD_SY + BOARD_CELL_SIZE * i + BOARD_CELL_SIZE / 2), colors.dark_gray);
		fonts.font_bold(coord_x[i]).draw(15, Arg::center(BOARD_SX + BOARD_CELL_SIZE * i + BOARD_CELL_SIZE / 2, BOARD_SY - BOARD_COORD_SIZE - 2), colors.dark_gray);
	}
	for (int i = 0; i < HW_M1; ++i) {
		Line(BOARD_SX + BOARD_CELL_SIZE * (i + 1), BOARD_SY, BOARD_SX + BOARD_CELL_SIZE * (i + 1), BOARD_SY + BOARD_CELL_SIZE * HW).draw(BOARD_CELL_FRAME_WIDTH, colors.dark_gray);
		Line(BOARD_SX, BOARD_SY + BOARD_CELL_SIZE * (i + 1), BOARD_SX + BOARD_CELL_SIZE * HW, BOARD_SY + BOARD_CELL_SIZE * (i + 1)).draw(BOARD_CELL_FRAME_WIDTH, colors.dark_gray);
	}
	Circle(BOARD_SX + 2 * BOARD_CELL_SIZE, BOARD_SY + 2 * BOARD_CELL_SIZE, BOARD_DOT_SIZE).draw(colors.dark_gray);
	Circle(BOARD_SX + 2 * BOARD_CELL_SIZE, BOARD_SY + 6 * BOARD_CELL_SIZE, BOARD_DOT_SIZE).draw(colors.dark_gray);
	Circle(BOARD_SX + 6 * BOARD_CELL_SIZE, BOARD_SY + 2 * BOARD_CELL_SIZE, BOARD_DOT_SIZE).draw(colors.dark_gray);
	Circle(BOARD_SX + 6 * BOARD_CELL_SIZE, BOARD_SY + 6 * BOARD_CELL_SIZE, BOARD_DOT_SIZE).draw(colors.dark_gray);
	RoundRect(BOARD_SX, BOARD_SY, BOARD_CELL_SIZE * HW, BOARD_CELL_SIZE * HW, BOARD_ROUND_DIAMETER).drawFrame(0, BOARD_ROUND_FRAME_WIDTH, colors.white);
	Flip flip;
	int board_arr[HW2];
	history_elem.board.translate_to_arr(board_arr, history_elem.player);
	for (int cell = 0; cell < HW2; ++cell) {
		int x = BOARD_SX + (cell % HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
		int y = BOARD_SY + (cell / HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
		if (board_arr[cell] == BLACK) {
			Circle(x, y, DISC_SIZE).draw(colors.black);
		}
		else if (board_arr[cell] == WHITE) {
			Circle(x, y, DISC_SIZE).draw(colors.white);
		}
	}
}

void draw_info(Colors colors, History_elem history_elem, Fonts fonts, Menu_elements menu_elements) {
	RoundRect round_rect{ INFO_SX, INFO_SY, INFO_WIDTH, INFO_HEIGHT, INFO_RECT_RADIUS };
	round_rect.drawFrame(INFO_RECT_THICKNESS, colors.white);
	if (history_elem.board.get_legal()) {
		fonts.font(Format(history_elem.board.n_discs() - 3) + language.get("info", "moves")).draw(13, Arg::topCenter(INFO_SX + INFO_WIDTH / 2, INFO_SY + 5));
		if (history_elem.player == BLACK) {
			fonts.font(language.get("info", "black")).draw(20, Arg::topCenter(INFO_SX + INFO_WIDTH / 2, INFO_SY + 22));
		}
		else {
			fonts.font(language.get("info", "white")).draw(20, Arg::topCenter(INFO_SX + INFO_WIDTH / 2, INFO_SY + 22));
		}
	}
	else {
		fonts.font(language.get("info", "game_end")).draw(20, Arg::topCenter(INFO_SX + INFO_WIDTH / 2, INFO_SY + 22));
	}
	Circle(INFO_SX + 70, INFO_SY + 60 + INFO_DISC_RADIUS, INFO_DISC_RADIUS).draw(colors.black);
	Circle(INFO_SX + INFO_WIDTH - 70, INFO_SY + 60 + INFO_DISC_RADIUS, INFO_DISC_RADIUS).draw(colors.white);
	int black_discs, white_discs;
	if (history_elem.player == BLACK) {
		black_discs = history_elem.board.count_player();
		white_discs = history_elem.board.count_opponent();
	}
	else {
		black_discs = history_elem.board.count_opponent();
		white_discs = history_elem.board.count_player();
	}
	fonts.font(black_discs).draw(20, Arg::leftCenter(INFO_SX + 100, INFO_SY + 60 + INFO_DISC_RADIUS));
	fonts.font(white_discs).draw(20, Arg::rightCenter(INFO_SX + INFO_WIDTH - 100, INFO_SY + 60 + INFO_DISC_RADIUS));
	Line(INFO_SX + INFO_WIDTH / 2, INFO_SY + 60, INFO_SX + INFO_WIDTH / 2, INFO_SY + 60 + INFO_DISC_RADIUS * 2).draw(2, colors.dark_gray);
	fonts.font(language.get("info", "opening_name") + U": " + Unicode::FromUTF8(history_elem.opening_name)).draw(12, Arg::topCenter(INFO_SX + INFO_WIDTH / 2, INFO_SY + 95));
	String level_info = language.get("common", "level") + U" " + Format(menu_elements.level) + U" (";
	if (menu_elements.level <= LIGHT_LEVEL) {
		level_info += language.get("info", "light");
	}
	else if (menu_elements.level <= STANDARD_MAX_LEVEL) {
		level_info += language.get("info", "standard");
	}
	else if (menu_elements.level <= PRAGMATIC_MAX_LEVEL) {
		level_info += language.get("info", "pragmatic");
	}
	else if (menu_elements.level <= ACCURATE_MAX_LEVEL) {
		level_info += language.get("info", "accurate");
	}
	else {
		level_info += language.get("info", "danger");
	}
	level_info += U")";
	fonts.font(level_info).draw(13, Arg::topCenter(INFO_SX + INFO_WIDTH / 2, INFO_SY + 115));
}

Umigame_result get_umigame(Board board, int player) {
	return umigame.get(&board, player);
}

class Main_scene : public App::Scene {
private:
	Graph graph;
	//Level_display level_display;
	Move_board_button_status move_board_button_status;
	AI_status ai_status;
	Button start_game_button;
	bool need_start_game_button;
	Umigame_status umigame_status;
	bool changing_scene;

public:
	Main_scene(const InitData& init) : IScene{ init } {
		cerr << "main scene loading" << endl;
		getData().menu = create_menu(&getData().menu_elements);
		graph.sx = GRAPH_SX;
		graph.sy = GRAPH_SY;
		graph.size_x = GRAPH_WIDTH;
		graph.size_y = GRAPH_HEIGHT;
		graph.resolution = GRAPH_RESOLUTION;
		if (getData().graph_resources.need_init) {
			getData().game_information.init();
			getData().graph_resources.init();
			getData().graph_resources.nodes[getData().graph_resources.put_mode].emplace_back(getData().history_elem);
		}
		start_game_button.init(START_GAME_BUTTON_SX, START_GAME_BUTTON_SY, START_GAME_BUTTON_WIDTH, START_GAME_BUTTON_HEIGHT, START_GAME_BUTTON_RADIUS, language.get("play", "start_game"), 15, getData().fonts.font, getData().colors.white, getData().colors.black);
		need_start_game_button_calculation();
		changing_scene = false;
		cerr << "main scene loaded" << endl;
	}

	void update() override {
		if (System::GetUserActions() & UserAction::CloseButtonClicked) {
			stop_calculating();
			changing_scene = true;
			changeScene(U"Close", SCENE_FADE_TIME);
			return;
		}

		Scene::SetBackground(getData().colors.green);

		// multi threading
		if (getData().menu_elements.n_threads != thread_pool.size()) {
			stop_calculating();
			thread_pool.resize(getData().menu_elements.n_threads);
			cerr << "thread pool resized to " << thread_pool.size() << endl;
			resume_calculating();
		}

		// init
		getData().graph_resources.delta = 0;

		// opening
		update_opening();

		// menu
		menu_game();
		menu_manipulate();
		menu_in_out();
		menu_book();
		menu_help();
		menu_language();

		// analyze
		if (ai_status.analyzing) {
			analyze_get_task();
		}

		// move
		bool ai_should_move =
			!need_start_game_button &&
			!getData().history_elem.board.is_end() && 
			getData().graph_resources.put_mode == GRAPH_MODE_NORMAL &&
			((getData().history_elem.player == BLACK && getData().menu_elements.ai_put_black) || (getData().history_elem.player == WHITE && getData().menu_elements.ai_put_white)) &&
			getData().history_elem.board.n_discs() == getData().graph_resources.nodes[GRAPH_MODE_NORMAL][getData().graph_resources.nodes[GRAPH_MODE_NORMAL].size() - 1].board.n_discs();
		if (need_start_game_button) {
			need_start_game_button_calculation();
			start_game_button.draw();
			if (start_game_button.clicked()) {
				need_start_game_button = false;
				stop_calculating();
				resume_calculating();
			}
		}
		if (!ai_status.analyzing) {
			if (ai_should_move) {
				ai_move();
			}
			else if (!getData().menu.active() && !need_start_game_button) {
				interact_move();
			}
		}

		bool graph_interact_ignore = ai_status.analyzing || ai_should_move;
		// transcript move
		if (!graph_interact_ignore && !getData().menu.active()) {
			interact_graph();
		}
		update_n_discs();

		// book modifying by right-clicking
		if (!ai_should_move && !need_start_game_button && getData().menu_elements.change_book_by_right_click) {
			change_book_by_right_click();
		}

		// board drawing
		draw_board(getData().fonts, getData().colors, getData().history_elem);

		// last move drawing
		if (getData().menu_elements.show_last_move) {
			draw_last_move();
		}

		// next move drawing
		if (getData().menu_elements.show_next_move) {
			draw_next_move();
		}

		uint64_t legal_ignore = 0ULL;

		// hint calculating & drawing
		bool hint_ignore = ai_should_move || ai_status.analyzing || need_start_game_button;
		if (!hint_ignore) {
			if (getData().menu_elements.use_disc_hint) {
				if (!ai_status.hint_calculating && ai_status.hint_level < getData().menu_elements.level) {
					hint_init_calculating();
				}
				hint_do_task();
				legal_ignore = draw_hint();
			}
		}

		// legal drawing
		if (getData().menu_elements.show_legal) {
			draw_legal(legal_ignore);
		}

		// umigame calculating / drawing
		if (getData().menu_elements.use_umigame_value && !hint_ignore) {
			if (umigame_status.umigame_calculated) {
				draw_umigame(legal_ignore);
			}
			else {
				calculate_umigame();
			}
		}

		// stable drawing
		if (getData().menu_elements.show_stable_discs) {
			draw_stable_discs();
		}

		// graph drawing
		graph.draw(getData().graph_resources.nodes[0], getData().graph_resources.nodes[1], getData().graph_resources.n_discs, getData().menu_elements.show_graph, getData().menu_elements.level);

		// level display drawing
		//level_display.draw(getData().menu_elements.level, getData().history_elem.board.n_discs());

		// info drawing
		draw_info(getData().colors, getData().history_elem, getData().fonts, getData().menu_elements);

		// opening on cell drawing
		if (getData().menu_elements.show_opening_on_cell) {
			draw_opening_on_cell();
		}

		// menu drawing
		getData().menu.draw();
	}

	void draw() const override {

	}

private:
	void reset_hint() {
		ai_status.hint_level = HINT_NOT_CALCULATING;
		ai_status.hint_available = false;
		ai_status.hint_calculating = false;
		ai_status.hint_task_stack.clear();
	}

	void reset_ai() {
		ai_status.ai_thinking = false;
	}

	void reset_analyze() {
		ai_status.analyzing = false;
		ai_status.analyze_task_stack.clear();
	}

	void reset_umigame() {
		umigame_status.umigame_calculated = false;
		umigame_status.umigame_calculating = false;
	}

	void stop_calculating() {
		cerr << "terminating calculation" << endl;
		global_searching = false;
		if (ai_status.ai_future.valid()) {
			ai_status.ai_future.get();
		}
		for (int i = 0; i < HW2; ++i) {
			if (ai_status.hint_future[i].valid()) {
				ai_status.hint_future[i].get();
			}
		}
		for (int i = 0; i < ANALYZE_SIZE; ++i) {
			if (ai_status.analyze_future[i].valid()) {
				ai_status.analyze_future[i].get();
			}
		}
		for (int i = 0; i < HW2; ++i) {
			if (umigame_status.umigame_future[i].valid()) {
				umigame_status.umigame_future[i].get();
			}
		}
		cerr << "calculation terminated" << endl;
		reset_ai();
		reset_hint();
		reset_analyze();
		reset_umigame();
		cerr << "reset all calculations" << endl;
	}

	void resume_calculating() {
		global_searching = true;
	}

	void menu_game() {
		if (getData().menu_elements.start_game) {
			stop_calculating();
			getData().history_elem.reset();
			getData().graph_resources.init();
			getData().graph_resources.nodes[getData().graph_resources.put_mode].emplace_back(getData().history_elem);
			getData().game_information.init();
			resume_calculating();
			need_start_game_button_calculation();
		}
		if (getData().menu_elements.analyze && !ai_status.ai_thinking && !ai_status.analyzing) {
			stop_calculating();
			init_analyze();
			resume_calculating();
		}
	}

	void menu_in_out() {
		if (getData().menu_elements.input_transcript) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Import_transcript", SCENE_FADE_TIME);
			return;
		}
		if (getData().menu_elements.input_board) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Import_board", SCENE_FADE_TIME);
			return;
		}
		if (getData().menu_elements.edit_board) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Edit_board", SCENE_FADE_TIME);
			return;
		}
		if (getData().menu_elements.input_game) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Import_game", SCENE_FADE_TIME);
			return;
		}
		if (getData().menu_elements.copy_transcript) {
			copy_transcript();
		}
		if (getData().menu_elements.save_game) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Export_game", SCENE_FADE_TIME);
			return;
		}
	}

	void menu_manipulate() {
		if (getData().menu_elements.stop_calculating) {
			stop_calculating();
			ai_status.hint_level = HINT_INF_LEVEL;
			resume_calculating();
		}
		if (!ai_status.analyzing) {
			if (getData().menu_elements.backward) {
				--getData().graph_resources.n_discs;
				getData().graph_resources.delta = -1;
			}
			if (getData().menu_elements.forward) {
				++getData().graph_resources.n_discs;
				getData().graph_resources.delta = 1;
			}
		}
		if (getData().menu_elements.convert_180) {
			stop_calculating();
			getData().history_elem.board.board_rotate_180();
			if (0 <= getData().history_elem.policy && getData().history_elem.policy < HW2) {
				getData().history_elem.policy = HW2_M1 - getData().history_elem.policy;
			}
			if (0 <= getData().history_elem.next_policy && getData().history_elem.next_policy < HW2) {
				getData().history_elem.next_policy = HW2_M1 - getData().history_elem.next_policy;
			}
			for (int i = 0; i < 2; ++i) {
				for (int j = 0; j < (int)getData().graph_resources.nodes[i].size(); ++j) {
					getData().graph_resources.nodes[i][j].board.board_rotate_180();
					if (0 <= getData().graph_resources.nodes[i][j].policy && getData().graph_resources.nodes[i][j].policy < HW2) {
						getData().graph_resources.nodes[i][j].policy = HW2_M1 - getData().graph_resources.nodes[i][j].policy;
					}
					if (0 <= getData().graph_resources.nodes[i][j].next_policy && getData().graph_resources.nodes[i][j].next_policy < HW2) {
						getData().graph_resources.nodes[i][j].next_policy = HW2_M1 - getData().graph_resources.nodes[i][j].next_policy;
					}
				}
			}
			resume_calculating();
		}
		if (getData().menu_elements.convert_blackline) {
			stop_calculating();
			getData().history_elem.board.board_black_line_mirror();
			if (0 <= getData().history_elem.policy && getData().history_elem.policy < HW2) {
				int x = getData().history_elem.policy % HW;
				int y = getData().history_elem.policy / HW;
				getData().history_elem.policy = (HW_M1 - x) * HW + (HW_M1 - y);
			}
			if (0 <= getData().history_elem.next_policy && getData().history_elem.next_policy < HW2) {
				int x = getData().history_elem.next_policy % HW;
				int y = getData().history_elem.next_policy / HW;
				getData().history_elem.next_policy = (HW_M1 - x) * HW + (HW_M1 - y);
			}
			for (int i = 0; i < 2; ++i) {
				for (int j = 0; j < (int)getData().graph_resources.nodes[i].size(); ++j) {
					getData().graph_resources.nodes[i][j].board.board_black_line_mirror();
					if (0 <= getData().graph_resources.nodes[i][j].policy && getData().graph_resources.nodes[i][j].policy < HW2) {
						int x = getData().graph_resources.nodes[i][j].policy % HW;
						int y = getData().graph_resources.nodes[i][j].policy / HW;
						getData().graph_resources.nodes[i][j].policy = (HW_M1 - x) * HW + (HW_M1 - y);
					}
					if (0 <= getData().graph_resources.nodes[i][j].next_policy && getData().graph_resources.nodes[i][j].next_policy < HW2) {
						int x = getData().graph_resources.nodes[i][j].next_policy % HW;
						int y = getData().graph_resources.nodes[i][j].next_policy / HW;
						getData().graph_resources.nodes[i][j].next_policy = (HW_M1 - x) * HW + (HW_M1 - y);
					}
				}
			}
			resume_calculating();
		}
		if (getData().menu_elements.convert_whiteline) {
			stop_calculating();
			getData().history_elem.board.board_white_line_mirror();
			if (0 <= getData().history_elem.policy && getData().history_elem.policy < HW2) {
				int x = getData().history_elem.policy % HW;
				int y = getData().history_elem.policy / HW;
				getData().history_elem.policy = x * HW + y;
			}
			if (0 <= getData().history_elem.next_policy && getData().history_elem.next_policy < HW2) {
				int x = getData().history_elem.next_policy % HW;
				int y = getData().history_elem.next_policy / HW;
				getData().history_elem.next_policy = x * HW + y;
			}
			for (int i = 0; i < 2; ++i) {
				for (int j = 0; j < (int)getData().graph_resources.nodes[i].size(); ++j) {
					getData().graph_resources.nodes[i][j].board.board_white_line_mirror();
					if (0 <= getData().graph_resources.nodes[i][j].policy && getData().graph_resources.nodes[i][j].policy < HW2) {
						int x = getData().graph_resources.nodes[i][j].policy % HW;
						int y = getData().graph_resources.nodes[i][j].policy / HW;
						getData().graph_resources.nodes[i][j].policy = x * HW + y;
					}
					if (0 <= getData().graph_resources.nodes[i][j].next_policy && getData().graph_resources.nodes[i][j].next_policy < HW2) {
						int x = getData().graph_resources.nodes[i][j].next_policy % HW;
						int y = getData().graph_resources.nodes[i][j].next_policy / HW;
						getData().graph_resources.nodes[i][j].next_policy = x * HW + y;
					}
				}
			}
			resume_calculating();
		}
	}

	void menu_book() {
		if (getData().menu_elements.book_start_learn) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Learn_book", SCENE_FADE_TIME);
			return;
		}
		if (getData().menu_elements.book_import) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Import_book", SCENE_FADE_TIME);
			return;
		}
		if (getData().menu_elements.book_reference) {
			stop_calculating();
			resume_calculating();
			changing_scene = true;
			changeScene(U"Refer_book", SCENE_FADE_TIME);
			return;
		}
	}

	void menu_help() {
		if (getData().menu_elements.website) {
			System::LaunchBrowser(U"https://www.egaroucid-app.nyanyan.dev/");
		}
		if (getData().menu_elements.bug_report) {
			System::LaunchBrowser(U"https://docs.google.com/forms/d/e/1FAIpQLSd6ML1T1fc707luPEefBXuImMnlM9cQP8j-YHKiSyFoS-8rmQ/viewform?usp=sf_link");
		}
		if (getData().menu_elements.license_egaroucid) {
			//System::LaunchBrowser(U"LICENSE");
			System::LaunchBrowser(U"https://github.com/Nyanyan/Egaroucid/blob/main/LICENSE");
		}
		if (getData().menu_elements.license_siv3d) {
			LicenseManager::ShowInBrowser();
		}
	}

	void menu_language() {
		for (int i = 0; i < (int)getData().resources.language_names.size(); ++i) {
			if (getData().menu_elements.languages[i]) {
				if (getData().settings.lang_name != getData().resources.language_names[i]) {
					string lang_file = "resources/languages/" + getData().resources.language_names[i] + ".json";
					if (!language.init(lang_file)) {
						cerr << "language setting error" << endl;
						lang_file = "resources/languages/" + getData().settings.lang_name + ".json";
						language.init(lang_file);
					}
					else if (!opening_init(getData().resources.language_names[i])) {
						cerr << "opening setting error" << endl;
						opening_init(getData().settings.lang_name);
					}
					else {
						getData().settings.lang_name = getData().resources.language_names[i];
						getData().menu = create_menu(&getData().menu_elements);
						start_game_button.init(START_GAME_BUTTON_SX, START_GAME_BUTTON_SY, START_GAME_BUTTON_WIDTH, START_GAME_BUTTON_HEIGHT, START_GAME_BUTTON_RADIUS, language.get("play", "start_game"), 15, getData().fonts.font, getData().colors.white, getData().colors.black);
						re_calculate_openings();
					}
				}
			}
		}
	}

	void interact_graph() {
		getData().graph_resources.n_discs = graph.update_n_discs(getData().graph_resources.nodes[0], getData().graph_resources.nodes[1], getData().graph_resources.n_discs);
		if (!KeyLeft.pressed() && !KeyA.pressed()) {
			move_board_button_status.left_pushed = BUTTON_NOT_PUSHED;
		}
		if (!KeyRight.pressed() && !KeyD.pressed()) {
			move_board_button_status.right_pushed = BUTTON_NOT_PUSHED;
		}

		if (MouseX1.down() || KeyLeft.down() || KeyA.down() || (move_board_button_status.left_pushed != BUTTON_NOT_PUSHED && tim() - move_board_button_status.left_pushed >= BUTTON_LONG_PRESS_THRESHOLD)) {
			--getData().graph_resources.n_discs;
			getData().graph_resources.delta = -1;
			if (KeyLeft.down() || KeyA.down()) {
				move_board_button_status.left_pushed = tim();
			}
		}
		else if (MouseX2.down() || KeyRight.down() || KeyD.down() || (move_board_button_status.right_pushed != BUTTON_NOT_PUSHED && tim() - move_board_button_status.right_pushed >= BUTTON_LONG_PRESS_THRESHOLD)) {
			++getData().graph_resources.n_discs;
			getData().graph_resources.delta = 1;
			if (KeyRight.down() || KeyD.down()) {
				move_board_button_status.right_pushed = tim();
			}
		}
	}

	void update_n_discs() {
		int max_n_discs = getData().graph_resources.nodes[getData().graph_resources.put_mode].back().board.n_discs();
		getData().graph_resources.n_discs = min(getData().graph_resources.n_discs, max_n_discs);
		int min_n_discs = getData().graph_resources.nodes[GRAPH_MODE_NORMAL][0].board.n_discs();
		if (getData().graph_resources.nodes[GRAPH_MODE_INSPECT].size()) {
			min_n_discs = min(min_n_discs, getData().graph_resources.nodes[GRAPH_MODE_INSPECT][0].board.n_discs());
		}
		getData().graph_resources.n_discs = max(getData().graph_resources.n_discs, min_n_discs);
		if (getData().graph_resources.put_mode == GRAPH_MODE_INSPECT && getData().graph_resources.nodes[GRAPH_MODE_INSPECT].size()) {
			if (getData().graph_resources.n_discs < getData().graph_resources.nodes[GRAPH_MODE_INSPECT][0].board.n_discs()) {
				getData().graph_resources.put_mode = GRAPH_MODE_NORMAL;
				getData().graph_resources.nodes[1].clear();
			}
		}
		int node_idx = getData().graph_resources.node_find(getData().graph_resources.put_mode, getData().graph_resources.n_discs);
		if (node_idx == -1 && getData().graph_resources.put_mode == GRAPH_MODE_INSPECT) {
			getData().graph_resources.nodes[GRAPH_MODE_INSPECT].clear();
			int node_idx_0 = getData().graph_resources.node_find(GRAPH_MODE_NORMAL, getData().graph_resources.n_discs);
			if (node_idx_0 == -1) {
				cerr << "history vector element not found 0" << endl;
				return;
			}
			getData().graph_resources.nodes[GRAPH_MODE_INSPECT].emplace_back(getData().graph_resources.nodes[GRAPH_MODE_NORMAL][node_idx_0]);
			node_idx = getData().graph_resources.node_find(getData().graph_resources.put_mode, getData().graph_resources.n_discs);
		}
		while (node_idx == -1) {
			//cerr << "history vector element not found 1" << endl;
			getData().graph_resources.n_discs += getData().graph_resources.delta;
			node_idx = getData().graph_resources.node_find(GRAPH_MODE_NORMAL, getData().graph_resources.n_discs);
		}
		if (getData().history_elem.board != getData().graph_resources.nodes[getData().graph_resources.put_mode][node_idx].board) {
			stop_calculating();
			resume_calculating();
		}
		getData().history_elem = getData().graph_resources.nodes[getData().graph_resources.put_mode][node_idx];
	}

	void move_processing(int_fast8_t cell) {
		int parent_idx = getData().graph_resources.node_find(getData().graph_resources.put_mode, getData().history_elem.board.n_discs());
		if (parent_idx != -1) {
			if (getData().graph_resources.nodes[getData().graph_resources.put_mode][parent_idx].next_policy == HW2_M1 - cell && parent_idx + 1 < (int)getData().graph_resources.nodes[getData().graph_resources.put_mode].size()) {
				++getData().graph_resources.n_discs;
				return;
			}
			getData().graph_resources.nodes[getData().graph_resources.put_mode][parent_idx].next_policy = HW2_M1 - cell;
			while (getData().graph_resources.nodes[getData().graph_resources.put_mode].size() > parent_idx + 1) {
				getData().graph_resources.nodes[getData().graph_resources.put_mode].pop_back();
			}
		}
		Flip flip;
		calc_flip(&flip, &getData().history_elem.board, HW2_M1 - cell);
		getData().history_elem.board.move_board(&flip);
		getData().history_elem.policy = HW2_M1 - cell;
		getData().history_elem.next_policy = -1;
		getData().history_elem.v = GRAPH_IGNORE_VALUE;
		getData().history_elem.level = -1;
		getData().history_elem.player ^= 1;
		if (getData().history_elem.board.get_legal() == 0ULL) {
			getData().history_elem.board.pass();
			getData().history_elem.player ^= 1;
		}
		getData().graph_resources.nodes[getData().graph_resources.put_mode].emplace_back(getData().history_elem);
		getData().graph_resources.n_discs++;
		reset_hint();
		reset_umigame();
	}

	void interact_move() {
		uint64_t legal = getData().history_elem.board.get_legal();
		for (int_fast8_t cell = 0; cell < HW2; ++cell) {
			if (1 & (legal >> (HW2_M1 - cell))) {
				int x = cell % HW;
				int y = cell / HW;
				Rect cell_rect(BOARD_SX + x * BOARD_CELL_SIZE, BOARD_SY + y * BOARD_CELL_SIZE, BOARD_CELL_SIZE, BOARD_CELL_SIZE);
				if (cell_rect.leftClicked()) {
					if (getData().graph_resources.put_mode == GRAPH_MODE_NORMAL) {
						int parent_idx = getData().graph_resources.node_find(GRAPH_MODE_NORMAL, getData().history_elem.board.n_discs());
						if (parent_idx != -1) {
							bool go_to_inspection_mode =
								getData().history_elem.board.n_discs() != getData().graph_resources.nodes[GRAPH_MODE_NORMAL][getData().graph_resources.nodes[GRAPH_MODE_NORMAL].size() - 1].board.n_discs() &&
								HW2_M1 - cell != getData().graph_resources.nodes[GRAPH_MODE_NORMAL][parent_idx].next_policy;
							if (go_to_inspection_mode) {
								getData().graph_resources.put_mode = GRAPH_MODE_INSPECT;
							}
						}
					}
					stop_calculating();
					move_processing(cell);
					resume_calculating();
				}
			}
		}
	}

	void ai_move() {
		if (!changing_scene) {
			uint64_t legal = getData().history_elem.board.get_legal();
			if (!ai_status.ai_thinking) {
				if (legal) {
					stop_calculating();
					resume_calculating();
					ai_status.ai_future = async(launch::async, ai, getData().history_elem.board, getData().menu_elements.level, getData().menu_elements.use_book, true, true);
					ai_status.ai_thinking = true;
				}
			}
			else if (ai_status.ai_future.valid()) {
				if (ai_status.ai_future.wait_for(chrono::seconds(0)) == future_status::ready) {
					Search_result search_result = ai_status.ai_future.get();
					if (1 & (legal >> search_result.policy)) {
						int sgn = getData().history_elem.player == 0 ? 1 : -1;
						getData().graph_resources.nodes[getData().graph_resources.put_mode].back().v = sgn * search_result.value;
						getData().graph_resources.nodes[getData().graph_resources.put_mode].back().level = getData().menu_elements.level;
						move_processing(HW2_M1 - search_result.policy);
					}
					ai_status.ai_thinking = false;
				}
			}
		}
	}

	void update_opening() {
		string new_opening = opening.get(getData().history_elem.board, getData().history_elem.player ^ 1);
		if (new_opening.size() && getData().history_elem.opening_name != new_opening) {
			getData().history_elem.opening_name = new_opening;
			int node_idx = getData().graph_resources.node_find(getData().graph_resources.put_mode, getData().graph_resources.n_discs);
			if (node_idx == -1) {
				cerr << "history vector element not found 2" << endl;
				return;
			}
			getData().graph_resources.nodes[getData().graph_resources.put_mode][node_idx].opening_name = new_opening;
		}
	}

	Menu create_menu(Menu_elements* menu_elements) {
		Menu menu;
		menu_title title;
		menu_elem menu_e, side_menu, side_side_menu;



		title.init(language.get("play", "game"));

		menu_e.init_button(language.get("play", "new_game"), &menu_elements->start_game);
		title.push(menu_e);
		menu_e.init_button(language.get("play", "analyze"), &menu_elements->analyze);
		title.push(menu_e);

		menu.push(title);




		title.init(language.get("settings", "settings"));

		menu_e.init_check(language.get("ai_settings", "use_book"), &menu_elements->use_book, menu_elements->use_book);
		title.push(menu_e);
		menu_e.init_bar(language.get("ai_settings", "level"), &menu_elements->level, menu_elements->level, 0, 60);
		title.push(menu_e);
		menu_e.init_bar(language.get("settings", "thread", "thread"), &menu_elements->n_threads, menu_elements->n_threads, 1, 32);
		title.push(menu_e);

		menu_e.init_check(language.get("settings", "play", "ai_put_black"), &menu_elements->ai_put_black, menu_elements->ai_put_black);
		title.push(menu_e);
		menu_e.init_check(language.get("settings", "play", "ai_put_white"), &menu_elements->ai_put_white, menu_elements->ai_put_white);
		title.push(menu_e);

		menu.push(title);




		title.init(language.get("display", "display"));

		menu_e.init_button(language.get("display", "cell", "display_on_cell"), &menu_elements->dummy);
		side_menu.init_check(language.get("display", "cell", "legal"), &menu_elements->show_legal, menu_elements->show_legal);
		menu_e.push(side_menu);
		side_menu.init_check(language.get("display", "cell", "disc_value"), &menu_elements->use_disc_hint, menu_elements->use_disc_hint);
		side_side_menu.init_bar(language.get("display", "cell", "disc_value_number"), &menu_elements->n_disc_hint, menu_elements->n_disc_hint, 1, SHOW_ALL_HINT);
		side_menu.push(side_side_menu);
		menu_e.push(side_menu);
		side_menu.init_check(language.get("display", "cell", "umigame_value"), &menu_elements->use_umigame_value, menu_elements->use_umigame_value);
		menu_e.push(side_menu);
		side_menu.init_check(language.get("display", "cell", "opening"), &menu_elements->show_opening_on_cell, menu_elements->show_opening_on_cell);
		menu_e.push(side_menu);
		title.push(menu_e);

		menu_e.init_button(language.get("display", "disc", "display_on_disc"), &menu_elements->dummy);
		side_menu.init_check(language.get("display", "disc", "last_move"), &menu_elements->show_last_move, menu_elements->show_last_move);
		menu_e.push(side_menu);
		side_menu.init_check(language.get("display", "disc", "next_move"), &menu_elements->show_next_move, menu_elements->show_next_move);
		menu_e.push(side_menu);
		side_menu.init_check(language.get("display", "disc", "stable"), &menu_elements->show_stable_discs, menu_elements->show_stable_discs);
		menu_e.push(side_menu);
		title.push(menu_e);

		menu_e.init_check(language.get("display", "graph"), &menu_elements->show_graph, menu_elements->show_graph);
		title.push(menu_e);
		menu_e.init_check(language.get("display", "log"), &menu_elements->show_log, menu_elements->show_log);
		title.push(menu_e);

		menu.push(title);




		title.init(language.get("operation", "operation"));

		menu_e.init_button(language.get("operation", "stop_calculating"), &menu_elements->stop_calculating);
		title.push(menu_e);

		menu_e.init_button(language.get("operation", "forward"), &menu_elements->forward);
		title.push(menu_e);
		menu_e.init_button(language.get("operation", "backward"), &menu_elements->backward);
		title.push(menu_e);

		menu_e.init_button(language.get("operation", "convert", "convert"), &menu_elements->dummy);
		side_menu.init_button(language.get("operation", "convert", "vertical"), &menu_elements->convert_180);
		menu_e.push(side_menu);
		side_menu.init_button(language.get("operation", "convert", "black_line"), &menu_elements->convert_blackline);
		menu_e.push(side_menu);
		side_menu.init_button(language.get("operation", "convert", "white_line"), &menu_elements->convert_whiteline);
		menu_e.push(side_menu);
		title.push(menu_e);

		menu.push(title);



		title.init(language.get("in_out", "in_out"));

		menu_e.init_button(language.get("in_out", "in"), &menu_elements->dummy);
		side_menu.init_button(language.get("in_out", "input_transcript"), &menu_elements->input_transcript);
		menu_e.push(side_menu);
		side_menu.init_button(language.get("in_out", "input_board"), &menu_elements->input_board);
		menu_e.push(side_menu);
		side_menu.init_button(language.get("in_out", "edit_board"), &menu_elements->edit_board);
		menu_e.push(side_menu);
		side_menu.init_button(language.get("in_out", "input_game"), &menu_elements->input_game);
		menu_e.push(side_menu);
		title.push(menu_e);

		menu_e.init_button(language.get("in_out", "out"), &menu_elements->dummy);
		side_menu.init_button(language.get("in_out", "output_transcript"), &menu_elements->copy_transcript);
		menu_e.push(side_menu);
		side_menu.init_button(language.get("in_out", "output_game"), &menu_elements->save_game);
		menu_e.push(side_menu);
		title.push(menu_e);

		menu.push(title);




		title.init(language.get("book", "book"));

		menu_e.init_check(language.get("book", "right_click_to_modify"), &menu_elements->change_book_by_right_click, menu_elements->change_book_by_right_click);
		title.push(menu_e);
		menu_e.init_button(language.get("book", "merge"), &menu_elements->book_import);
		title.push(menu_e);
		menu_e.init_button(language.get("book", "book_reference"), &menu_elements->book_reference);
		title.push(menu_e);
		menu_e.init_button(language.get("book", "settings"), &menu_elements->dummy);
		side_menu.init_bar(language.get("book", "depth"), &menu_elements->book_learn_depth, menu_elements->book_learn_depth, 0, 60);
		menu_e.push(side_menu);
		side_menu.init_bar(language.get("book", "accept"), &menu_elements->book_learn_error, menu_elements->book_learn_error, 0, 32);
		menu_e.push(side_menu);
		side_menu.init_check(language.get("book", "ignore"), &menu_elements->ignore_book, menu_elements->ignore_book);
		menu_e.push(side_menu);
		title.push(menu_e);
		menu_e.init_button(language.get("book", "start_learn"), &menu_elements->book_start_learn);
		title.push(menu_e);

		menu.push(title);




		title.init(language.get("help", "help"));
		menu_e.init_button(language.get("help", "website"), &menu_elements->website);
		title.push(menu_e);
		menu_e.init_button(language.get("help", "bug_report"), &menu_elements->bug_report);
		title.push(menu_e);
		menu_e.init_check(language.get("help", "auto_update_check"), &menu_elements->auto_update_check, menu_elements->auto_update_check);
		title.push(menu_e);
		menu_e.init_button(language.get("help", "license_egaroucid"), &menu_elements->license_egaroucid);
		title.push(menu_e);
		menu_e.init_button(language.get("help", "license_siv3d"), &menu_elements->license_siv3d);
		title.push(menu_e);
		menu.push(title);





		title.init(U"Language");
		for (int i = 0; i < (int)getData().resources.language_names.size(); ++i) {
			menu_e.init_radio(language_name.get(getData().resources.language_names[i]), &menu_elements->languages[i], menu_elements->languages[i]);
			title.push(menu_e);
		}
		menu.push(title);




		menu.init(0, 0, 12, getData().fonts.font, getData().resources.checkbox, getData().resources.unchecked);
		return menu;
	}

	void draw_legal(uint64_t ignore) {
		uint64_t legal = getData().history_elem.board.get_legal();
		for (int cell = 0; cell < HW2; ++cell) {
			int x = BOARD_SX + (cell % HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
			int y = BOARD_SY + (cell / HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
			if (1 & (legal >> (HW2_M1 - cell))) {
				if ((1 & (ignore >> (HW2_M1 - cell))) == 0)
					Circle(x, y, LEGAL_SIZE).draw(getData().colors.cyan);
			}
		}
	}

	void draw_last_move() {
		if (getData().history_elem.policy != -1) {
			int x = BOARD_SX + (HW_M1 - getData().history_elem.policy % HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
			int y = BOARD_SY + (HW_M1 - getData().history_elem.policy / HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
			Circle(x, y, LEGAL_SIZE).draw(getData().colors.red);
		}
	}

	void draw_next_move() {
		if (0 <= getData().history_elem.next_policy && getData().history_elem.next_policy <= HW2) {
			uint64_t legal = getData().history_elem.board.get_legal();
			if (1 & (legal >> getData().history_elem.next_policy)) {
				int x = BOARD_SX + (HW_M1 - getData().history_elem.next_policy % HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
				int y = BOARD_SY + (HW_M1 - getData().history_elem.next_policy / HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
				if (getData().history_elem.player == WHITE) {
					Circle(x, y, DISC_SIZE).draw(ColorF(getData().colors.white, 0.2));
				}
				else {
					Circle(x, y, DISC_SIZE).draw(ColorF(getData().colors.black, 0.2));
				}
			}
		}
	}

	uint64_t draw_hint() {
		uint64_t res = 0ULL;
		if (ai_status.hint_available) {
			vector<Hint_info> hint_infos;
			for (int cell = 0; cell < HW2; ++cell) {
				if (ai_status.hint_use_stable[cell]) {
					Hint_info hint_info;
					hint_info.value = ai_status.hint_values_stable[cell];
					hint_info.cell = cell;
					hint_info.type = ai_status.hint_types_stable[cell];
					hint_infos.emplace_back(hint_info);
				}
			}
			sort(hint_infos.begin(), hint_infos.end(), compare_hint_info);
			if (hint_infos.size()) {
				int sgn = getData().history_elem.player == 0 ? 1 : -1;
				int node_idx = getData().graph_resources.node_find(getData().graph_resources.put_mode, getData().graph_resources.n_discs);
				if (node_idx != -1) {
					if (getData().graph_resources.nodes[getData().graph_resources.put_mode][node_idx].level < hint_infos[0].type) {
						getData().graph_resources.nodes[getData().graph_resources.put_mode][node_idx].v = sgn * (int)round(hint_infos[0].value);
						getData().graph_resources.nodes[getData().graph_resources.put_mode][node_idx].level = hint_infos[0].type;
					}
				}
			}
			int n_disc_hint = min((int)hint_infos.size(), getData().menu_elements.n_disc_hint);
			for (int i = 0; i < n_disc_hint; ++i) {
				int sx = BOARD_SX + (hint_infos[i].cell % HW) * BOARD_CELL_SIZE;
				int sy = BOARD_SY + (hint_infos[i].cell / HW) * BOARD_CELL_SIZE;
				Color color = getData().colors.white;
				if (hint_infos[i].value == hint_infos[0].value) {
					color = getData().colors.cyan;
				}
				getData().fonts.font_bold((int)round(hint_infos[i].value)).draw(17, sx + 2, sy, color);
				if (hint_infos[i].type == HINT_TYPE_BOOK) {
					getData().fonts.font(U"book").draw(11, sx + 2, sy + 18, color);
				}
				else if (hint_infos[i].type > HINT_MAX_LEVEL) {
					getData().fonts.font(Format(hint_infos[i].type) + U"%").draw(11, sx + 2, sy + 18, color);
				}
				else {
					getData().fonts.font(U"Lv." + Format(hint_infos[i].type)).draw(11, sx + 2, sy + 18, color);
				}
				res |= 1ULL << (HW2_M1 - hint_infos[i].cell);
			}
		}
		return res;
	}

	void draw_stable_discs() {
		uint64_t stable = calc_stability_bits(&getData().history_elem.board);
		for (int cell = 0; cell < HW2; ++cell) {
			int x = BOARD_SX + (cell % HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
			int y = BOARD_SY + (cell / HW) * BOARD_CELL_SIZE + BOARD_CELL_SIZE / 2;
			if (1 & (stable >> (HW2_M1 - cell))) {
				Circle(x, y, STABLE_SIZE).draw(getData().colors.burlywood);
			}
		}
	}

	void draw_opening_on_cell() {
		uint64_t legal = getData().history_elem.board.get_legal();
		for (uint_fast8_t cell = first_bit(&legal); legal; cell = next_bit(&legal)) {
			int x = HW_M1 - cell % HW;
			int y = HW_M1 - cell / HW;
			Rect cell_rect(BOARD_SX + x * BOARD_CELL_SIZE, BOARD_SY + y * BOARD_CELL_SIZE, BOARD_CELL_SIZE, BOARD_CELL_SIZE);
			if (cell_rect.mouseOver()) {
				Flip flip;
				calc_flip(&flip, &getData().history_elem.board, cell);
				string openings = opening_many.get(getData().history_elem.board.move_copy(&flip), getData().history_elem.player);
				if (openings.size()) {
					String opening_name = U" " + Unicode::FromUTF8(openings).replace(U"|", U" \n ") + U" ";
					Vec2 pos = Cursor::Pos();
					pos.x += 20;
					RectF background_rect = getData().fonts.font_bold(opening_name).region(15, pos);
					background_rect.draw(getData().colors.white);
					getData().fonts.font_bold(opening_name).draw(15, pos, getData().colors.black);
				}
			}
		}
	}

	bool same_level(int l1, int l2, int n_moves) {
		bool m1, m2;
		int d1, d2;
		bool u1, u2;
		double t1, t2;
		get_level(l1, n_moves, &m1, &d1, &u1, &t1);
		get_level(l2, n_moves, &m2, &d2, &u2, &t2);
		return m1 == m2 && d1 == d2 && u1 == u2 && t1 == t2;
	}

	void hint_init_calculating() {
		uint64_t legal = getData().history_elem.board.get_legal();
		if (ai_status.hint_level == HINT_NOT_CALCULATING) {
			stop_calculating();
			resume_calculating();
			for (int cell = 0; cell < HW2; ++cell) {
				ai_status.hint_values[cell] = HINT_INIT_VALUE;
				ai_status.hint_use[cell] = (bool)(1 & (legal >> (HW2_M1 - cell)));
				ai_status.hint_types[cell] = HINT_NOT_CALCULATING;
			}
			for (int cell = 0; cell < HW2; ++cell) {
				ai_status.hint_use_stable[cell] = ai_status.hint_use[cell];
				ai_status.hint_values_stable[cell] = ai_status.hint_values[cell];
				ai_status.hint_types_stable[cell] = ai_status.hint_types[cell];
			}
		}
		else {
			ai_status.hint_available = true;
		}
		int before_level = ai_status.hint_level;
		int n_moves = getData().history_elem.board.n_discs() - 4;
		++ai_status.hint_level;
		if (ai_status.hint_level > 1) {
			while (same_level(before_level, ai_status.hint_level, n_moves) && ai_status.hint_level <= getData().menu_elements.level) {
				++ai_status.hint_level;
			}
		}
		if (ai_status.hint_level <= getData().menu_elements.level) {
			vector<pair<int, int>> value_cells;
			for (int cell = 0; cell < HW2; ++cell) {
				if (ai_status.hint_use[cell]) {
					value_cells.emplace_back(make_pair(ai_status.hint_values[cell], cell));
				}
			}
			sort(value_cells.begin(), value_cells.end(), compare_value_cell);
			int n_legal = pop_count_ull(legal);
			int hint_adoption_threshold = getData().menu_elements.n_disc_hint + max(1, n_legal * (getData().menu_elements.level - ai_status.hint_level) / getData().menu_elements.level);
			hint_adoption_threshold = min(hint_adoption_threshold, (int)value_cells.size());
			ai_status.hint_task_stack.clear();
			Board board;
			Flip flip;
			int next_task_size = 0;
			int idx = 0;
			for (pair<int, int>& value_cell : value_cells) {
				if (idx++ >= hint_adoption_threshold) {
					break;
				}
				if (ai_status.hint_types[value_cell.second] != HINT_TYPE_BOOK) {
					++next_task_size;
				}
			}
			ai_status.hint_use_multi_thread = next_task_size < getData().menu_elements.n_threads;
			
			if (ai_status.hint_level <= 10) {
				ai_status.hint_use_multi_thread = false;
			}
			
			//ai_status.hint_use_multi_thread = true;
			idx = 0;
			for (pair<int, int>& value_cell : value_cells) {
				if (idx++ >= hint_adoption_threshold) {
					break;
				}
				if (ai_status.hint_types[value_cell.second] != HINT_TYPE_BOOK) {
					board = getData().history_elem.board;
					calc_flip(&flip, &board, (uint_fast8_t)(HW2_M1 - value_cell.second));
					board.move_board(&flip);
					ai_status.hint_task_stack.emplace_back(make_pair(value_cell.second, bind(ai_hint, board, ai_status.hint_level, getData().menu_elements.use_book, ai_status.hint_use_multi_thread, false)));
				}
			}
			ai_status.hint_n_doing_tasks = 0;
			ai_status.hint_calculating = true;
			cerr << "hint search level " << ai_status.hint_level << " n_tasks " << ai_status.hint_task_stack.size() << " multi_threading " << ai_status.hint_use_multi_thread << endl;
		}
	}

	void hint_do_task() {
		if (!changing_scene) {
			if (ai_status.hint_n_doing_tasks > 0) {
				for (int cell = 0; cell < HW2; ++cell) {
					if (ai_status.hint_future[cell].valid()) {
						if (ai_status.hint_future[cell].wait_for(chrono::seconds(0)) == future_status::ready) {
							Search_result search_result = ai_status.hint_future[cell].get();
							if (ai_status.hint_values[cell] == HINT_INIT_VALUE || search_result.is_end_search || search_result.depth == SEARCH_BOOK) {
								ai_status.hint_values[cell] = -search_result.value;
							}
							else {
								ai_status.hint_values[cell] -= 1.2 * search_result.value;
								ai_status.hint_values[cell] /= 2.2;
							}
							if (search_result.depth == SEARCH_BOOK) {
								ai_status.hint_types[cell] = HINT_TYPE_BOOK;
							}
							else if (search_result.is_end_search) {
								ai_status.hint_types[cell] = search_result.probability;
							}
							else {
								ai_status.hint_types[cell] = ai_status.hint_level;
							}
							--ai_status.hint_n_doing_tasks;
						}
					}
				}
			}
			if (ai_status.hint_task_stack.size() == 0 && ai_status.hint_n_doing_tasks == 0) {
				for (int cell = 0; cell < HW2; ++cell) {
					ai_status.hint_use_stable[cell] = ai_status.hint_use[cell];
					ai_status.hint_values_stable[cell] = ai_status.hint_values[cell];
					ai_status.hint_types_stable[cell] = ai_status.hint_types[cell];
				}
				ai_status.hint_calculating = false;
			}
			else if (ai_status.hint_task_stack.size()) {
				int loop_time = min((int)ai_status.hint_task_stack.size(), getData().menu_elements.n_threads - ai_status.hint_n_doing_tasks);
				
				if (ai_status.hint_use_multi_thread) {
					if (max(1, getData().menu_elements.n_threads / HINT_SINGLE_TASK_N_THREAD) - ai_status.hint_n_doing_tasks > 0) {
						loop_time = min((int)ai_status.hint_task_stack.size(), max(1, getData().menu_elements.n_threads / HINT_SINGLE_TASK_N_THREAD) - ai_status.hint_n_doing_tasks);
					}
					else {
						loop_time = 0;
					}
				}
				else {
					loop_time = min((int)ai_status.hint_task_stack.size(), getData().menu_elements.n_threads - ai_status.hint_n_doing_tasks);
				}
				
				if (loop_time > 0) {
					for (int i = 0; i < loop_time; ++i) {
						pair<int, function<Search_result()>> task = ai_status.hint_task_stack.back();
						ai_status.hint_task_stack.pop_back();
						ai_status.hint_future[task.first] = async(launch::async, task.second);
					}
					ai_status.hint_n_doing_tasks += loop_time;
				}
			}
		}
	}

	void init_analyze() {
		ai_status.analyze_task_stack.clear();
		int idx = 0;
		for (History_elem& node : getData().graph_resources.nodes[getData().graph_resources.put_mode]) {
			Analyze_info analyze_info;
			analyze_info.idx = idx++;
			analyze_info.sgn = node.player ? -1 : 1;
			analyze_info.board = node.board;
			ai_status.analyze_task_stack.emplace_back(make_pair(analyze_info, bind(ai, node.board, getData().menu_elements.level, getData().menu_elements.use_book, true, true)));
		}
		cerr << "analyze " << ai_status.analyze_task_stack.size() << " tasks" << endl;
		ai_status.analyzing = true;
		analyze_do_task();
	}

	void analyze_do_task() {
		if (!changing_scene) {
			if (ai_status.analyze_task_stack.size() == 0) {
				ai_status.analyzing = false;
				getData().history_elem = getData().graph_resources.nodes[getData().graph_resources.put_mode].back();
				getData().graph_resources.n_discs = getData().graph_resources.nodes[getData().graph_resources.put_mode].back().board.n_discs();
				return;
			}
			pair<Analyze_info, function<Search_result()>> task = ai_status.analyze_task_stack.back();
			ai_status.analyze_task_stack.pop_back();
			ai_status.analyze_future[task.first.idx] = async(launch::async, task.second);
			ai_status.analyze_sgn[task.first.idx] = task.first.sgn;
			getData().history_elem.board = task.first.board;
			getData().history_elem.policy = -1;
			getData().history_elem.next_policy = -1;
			getData().history_elem.player = task.first.sgn == 1 ? 0 : 1;
			getData().graph_resources.n_discs = getData().history_elem.board.n_discs();
		}
	}

	void analyze_get_task() {
		bool task_finished = false;
		for (int i = 0; i < ANALYZE_SIZE; ++i) {
			if (ai_status.analyze_future[i].valid()) {
				if (ai_status.analyze_future[i].wait_for(chrono::seconds(0)) == future_status::ready) {
					Search_result search_result = ai_status.analyze_future[i].get();
					int value = ai_status.analyze_sgn[i] * search_result.value;
					getData().graph_resources.nodes[getData().graph_resources.put_mode][i].v = value;
					getData().graph_resources.nodes[getData().graph_resources.put_mode][i].level = getData().menu_elements.level;
					task_finished = true;
				}
			}
		}
		if (task_finished) {
			analyze_do_task();
		}
	}

	void copy_transcript() {
		string transcript;
		int inspect_switch_n_discs = INF;
		if (getData().graph_resources.put_mode == 1) {
			if (getData().graph_resources.nodes[GRAPH_MODE_INSPECT].size()) {
				inspect_switch_n_discs = getData().graph_resources.nodes[GRAPH_MODE_INSPECT][0].board.n_discs();
			}
			else {
				cerr << "no node found in inspect mode" << endl;
			}
		}
		cerr << inspect_switch_n_discs << endl;
		for (History_elem& history_elem : getData().graph_resources.nodes[GRAPH_MODE_NORMAL]) {
			if (history_elem.board.n_discs() + 1 >= inspect_switch_n_discs || history_elem.board.n_discs() >= getData().history_elem.board.n_discs()) {
				break;
			}
			if (history_elem.next_policy != -1) {
				transcript += idx_to_coord(history_elem.next_policy);
			}
		}
		if (inspect_switch_n_discs != INF) {
			if (getData().graph_resources.nodes[GRAPH_MODE_INSPECT][0].policy != -1) {
				transcript += idx_to_coord(getData().graph_resources.nodes[GRAPH_MODE_INSPECT][0].policy);
			}
			for (History_elem& history_elem : getData().graph_resources.nodes[GRAPH_MODE_INSPECT]) {
				if (history_elem.board.n_discs() >= getData().history_elem.board.n_discs()) {
					break;
				}
				if (history_elem.next_policy != -1) {
					transcript += idx_to_coord(history_elem.next_policy);
				}
			}
		}
		cerr << transcript << endl;
		Clipboard::SetText(Unicode::Widen(transcript));
	}

	void need_start_game_button_calculation() {
		need_start_game_button =
			((getData().history_elem.player == BLACK && getData().menu_elements.ai_put_black) ||
			(getData().history_elem.player == WHITE && getData().menu_elements.ai_put_white)) &&
			getData().history_elem.board.n_discs() == getData().graph_resources.nodes[getData().graph_resources.put_mode].back().board.n_discs() &&
			!getData().history_elem.board.is_end();
	}

	void calculate_umigame() {
		if (!changing_scene) {
			uint64_t legal = getData().history_elem.board.get_legal();
			if (!umigame_status.umigame_calculating) {
				Board board = getData().history_elem.board;
				int n_player = getData().history_elem.player ^ 1;
				Flip flip;
				for (uint_fast8_t cell = first_bit(&legal); legal; cell = next_bit(&legal)) {
					calc_flip(&flip, &board, cell);
					board.move_board(&flip);
					umigame_status.umigame_future[cell] = async(launch::async, get_umigame, board, n_player);
					board.undo_board(&flip);
				}
				umigame_status.umigame_calculating = true;
				cerr << "start umigame calculation" << endl;
			}
			else {
				bool all_done = true;
				for (uint_fast8_t cell = first_bit(&legal); legal; cell = next_bit(&legal)) {
					if (umigame_status.umigame_future[cell].valid()) {
						if (umigame_status.umigame_future[cell].wait_for(chrono::seconds(0)) == future_status::ready) {
							umigame_status.umigame[cell] = umigame_status.umigame_future[cell].get();
						}
						else {
							all_done = false;
						}
					}
				}
				if (all_done) {
					umigame_status.umigame_calculated = all_done;
					cerr << "finish umigame calculation" << endl;
				}
			}
		}
	}

	void draw_umigame(uint64_t legal_ignore) {
		uint64_t legal = getData().history_elem.board.get_legal();
		for (uint_fast8_t cell = first_bit(&legal); legal; cell = next_bit(&legal)) {
			if (1 & (legal_ignore >> cell)) {
				int sx = BOARD_SX + ((HW2_M1 - cell) % HW) * BOARD_CELL_SIZE;
				int sy = BOARD_SY + ((HW2_M1 - cell) / HW) * BOARD_CELL_SIZE;
				if (umigame_status.umigame[cell].b != UMIGAME_UNDEFINED) {
					getData().fonts.font_heavy(umigame_status.umigame[cell].b).draw(13, Arg::bottomRight(sx + BOARD_CELL_SIZE - 3, sy + BOARD_CELL_SIZE - 17), getData().colors.black);
					getData().fonts.font_heavy(umigame_status.umigame[cell].w).draw(13, Arg::bottomRight(sx + BOARD_CELL_SIZE - 3, sy + BOARD_CELL_SIZE - 1), getData().colors.white);
				}
			}
		}
	}

	void change_book_by_right_click() {
		if (getData().book_information.changing != BOOK_CHANGE_NO_CELL) {
			getData().fonts.font(language.get("book", "changed_value") + U"(" + Unicode::Widen(idx_to_coord(getData().book_information.changing)) + U"): " + getData().book_information.val_str).draw(13, CHANGE_BOOK_INFO_SX, CHANGE_BOOK_INFO_SY, getData().colors.white);
			if (KeyEscape.down()) {
				getData().book_information.changing = BOOK_CHANGE_NO_CELL;
			}
			else if (Key0.down() || KeyNum0.down()) {
				if (getData().book_information.val_str != U"0" && getData().book_information.val_str != U"-") {
					getData().book_information.val_str += U"0";
				}
			}
			else if (Key1.down() || KeyNum1.down()) {
				getData().book_information.val_str += U"1";
			}
			else if (Key2.down() || KeyNum2.down()) {
				getData().book_information.val_str += U"2";
			}
			else if (Key3.down() || KeyNum3.down()) {
				getData().book_information.val_str += U"3";
			}
			else if (Key4.down() || KeyNum4.down()) {
				getData().book_information.val_str += U"4";
			}
			else if (Key5.down() || KeyNum5.down()) {
				getData().book_information.val_str += U"5";
			}
			else if (Key6.down() || KeyNum6.down()) {
				getData().book_information.val_str += U"6";
			}
			else if (Key7.down() || KeyNum7.down()) {
				getData().book_information.val_str += U"7";
			}
			else if (Key8.down() || KeyNum8.down()) {
				getData().book_information.val_str += U"8";
			}
			else if (Key9.down() || KeyNum9.down()) {
				getData().book_information.val_str += U"9";
			}
			else if (KeyMinus.down()) {
				if (getData().book_information.val_str == U"" || getData().book_information.val_str == U"-") {
					getData().book_information.val_str += U"-";
				}
			}
			else if (KeyBackspace.down()) {
				if (getData().book_information.val_str.size()) {
					getData().book_information.val_str.pop_back();
				}
			}
		}
		uint64_t legal = getData().history_elem.board.get_legal();
		for (uint_fast8_t cell = first_bit(&legal); legal; cell = next_bit(&legal)) {
			int x = HW_M1 - cell % HW;
			int y = HW_M1 - cell / HW;
			Rect cell_rect(BOARD_SX + x * BOARD_CELL_SIZE, BOARD_SY + y * BOARD_CELL_SIZE, BOARD_CELL_SIZE, BOARD_CELL_SIZE);
			int state = 0;
			if (cell_rect.rightClicked()) {
				state = 1;
			}
			else if (KeyEnter.down()) {
				state = 2;
			}
			if (state) {
				if (getData().book_information.changing == cell) {
					if (getData().book_information.val_str.size() == 0) {
						getData().book_information.changing = BOOK_CHANGE_NO_CELL;
					}
					else {
						int changed_book_value = ParseOr<int>(getData().book_information.val_str, CHANGE_BOOK_ERR);
						if (changed_book_value < -HW2 || HW2 < changed_book_value) {
							getData().book_information.val_str.clear();
							changed_book_value = CHANGE_BOOK_ERR;
						}
						if (changed_book_value != CHANGE_BOOK_ERR) {
							cerr << "new value " << changed_book_value << endl;
							Flip flip;
							calc_flip(&flip, &getData().history_elem.board, getData().book_information.changing);
							Board b = getData().history_elem.board.move_copy(&flip);
							book.change(getData().history_elem.board.move_copy(&flip), changed_book_value);
							getData().book_information.changed = true;
							getData().book_information.changing = BOOK_CHANGE_NO_CELL;
							getData().book_information.val_str.clear();
							stop_calculating();
							resume_calculating();
						}
						else {
							Flip flip;
							calc_flip(&flip, &getData().history_elem.board, getData().book_information.changing);
							Board b = getData().history_elem.board.move_copy(&flip);
							book.delete_elem(b);
							getData().book_information.changed = true;
							getData().book_information.changing = BOOK_CHANGE_NO_CELL;
							getData().book_information.val_str.clear();
							stop_calculating();
							resume_calculating();
						}
					}
				}
				else if (state == 1) {
					getData().book_information.val_str.clear();
					getData().book_information.changing = cell;
				}
			}
		}
	}

	void re_calculate_openings() {
		for (int i = 0; i < 2; ++i) {
			for (int j = 0; j < (int)getData().graph_resources.nodes[i].size(); ++j) {
				getData().graph_resources.nodes[i][j].opening_name.clear();
			}
		}
		getData().history_elem.opening_name.clear();
		string opening_name = "";
		for (int i = 0; i < (int)getData().graph_resources.nodes[0].size(); ++i) {
			string new_opening = opening.get(getData().graph_resources.nodes[0][i].board, getData().graph_resources.nodes[0][i].player ^ 1);
			if (new_opening.size()) {
				opening_name = new_opening;
			}
			getData().graph_resources.nodes[0][i].opening_name = opening_name;
		}
		if (getData().graph_resources.nodes[1].size()) {
			int fork_n_discs = getData().graph_resources.nodes[1][0].board.n_discs();
			int node_idx = getData().graph_resources.node_find(0, fork_n_discs);
			if (node_idx == -1) {
				cerr << "history vector element not found" << endl;
			}
			else {
				node_idx = max(0, node_idx - 1);
				opening_name = getData().graph_resources.nodes[0][node_idx].opening_name;
				for (int i = 0; i < (int)getData().graph_resources.nodes[1].size(); ++i) {
					string new_opening = opening.get(getData().graph_resources.nodes[1][i].board, getData().graph_resources.nodes[1][i].player ^ 1);
					if (new_opening.size()) {
						opening_name = new_opening;
					}
					getData().graph_resources.nodes[1][i].opening_name = opening_name;
				}
			}
		}
		getData().history_elem.opening_name = opening.get(getData().history_elem.board, getData().history_elem.player ^ 1);
		if (getData().history_elem.opening_name.size() == 0) {
			int now_node_idx = getData().graph_resources.node_find(getData().graph_resources.put_mode, getData().history_elem.board.n_discs());
			if (now_node_idx == -1) {
				cerr << "history vector element not found" << endl;
			}
			else {
				now_node_idx = max(0, now_node_idx - 1);
				getData().history_elem.opening_name = getData().graph_resources.nodes[getData().graph_resources.put_mode][now_node_idx].opening_name;
			}
		}
	}
};
